use std::net::SocketAddr;

use anyhow::Context;
use mad::Component;

use crate::{
    services::staging::{StagingEvent, StagingState},
    state::SharedState,
};

include!("gen/nodata.rs");

#[derive(Clone)]
pub struct GrpcServer {
    state: SharedState,
    host: SocketAddr,
}

impl GrpcServer {
    pub fn new(state: &SharedState, host: SocketAddr) -> Self {
        Self {
            state: state.clone(),
            host,
        }
    }
}

#[tonic::async_trait]
impl no_data_server::NoData for GrpcServer {
    async fn publish_event(
        &self,
        request: tonic::Request<PublishEventRequest>,
    ) -> std::result::Result<tonic::Response<PublishEventResponse>, tonic::Status> {
        let req = request.into_inner();

        tracing::debug!(
            topic = req.topic,
            key = req.key,
            value = std::str::from_utf8(&req.value).ok(),
            id = req.id,
            "handling event"
        );

        self.state.staging().publish(req).await.map_err(|e| {
            tracing::warn!(error = e.to_string(), "failed to handle ingest of data");
            tonic::Status::internal(e.to_string())
        })?;

        Ok(tonic::Response::new(PublishEventResponse {}))
    }

    async fn get_topics(
        &self,
        _request: tonic::Request<GetTopicsRequest>,
    ) -> std::result::Result<tonic::Response<GetTopicsResponse>, tonic::Status> {
        let topics = self.state.staging.get_topics().await.map_err(|e| {
            tracing::warn!(error = e.to_string(), "failed to get topics");
            tonic::Status::internal(e.to_string())
        })?;

        Ok(tonic::Response::new(GetTopicsResponse { topics }))
    }

    async fn get_keys(
        &self,
        request: tonic::Request<GetKeysRequest>,
    ) -> std::result::Result<tonic::Response<GetKeysResponse>, tonic::Status> {
        let req = request.into_inner();

        let keys = self.state.staging.get_keys(&req.topic).await.map_err(|e| {
            tracing::warn!(error = e.to_string(), "failed to get keys");
            tonic::Status::internal(e.to_string())
        })?;

        Ok(tonic::Response::new(GetKeysResponse { keys }))
    }
}

impl From<PublishEventRequest> for StagingEvent {
    fn from(value: PublishEventRequest) -> Self {
        Self {
            topic: value.topic,
            id: value.id,
            key: value.key,
        }
    }
}

#[axum::async_trait]
impl Component for GrpcServer {
    fn name(&self) -> Option<String> {
        Some("grpc_server".into())
    }

    async fn run(
        &self,
        _cancellation_token: tokio_util::sync::CancellationToken,
    ) -> Result<(), mad::MadError> {
        tracing::info!("grpc listening on: {}", self.host);

        tonic::transport::Server::builder()
            .add_service(no_data_server::NoDataServer::new(self.clone()))
            .serve(self.host)
            .await
            .context("grpc server failed")
            .map_err(mad::MadError::Inner)?;

        Ok(())
    }
}
