mod grpc;
mod http;
mod state;

mod services;

use std::net::SocketAddr;

use chrono::{Datelike, Timelike};
use clap::{Parser, Subcommand, ValueEnum};
use grpc::{GetKeysRequest, GetTopicsRequest, GrpcServer, PublishEventRequest};
use http::HttpServer;
use mad::Mad;
use state::SharedState;
use uuid::Uuid;

#[derive(Parser)]
#[command(author, version, about, long_about = None, subcommand_required = true)]
struct Command {
    #[command(subcommand)]
    command: Option<Commands>,
}

#[derive(Subcommand)]
enum Commands {
    Serve {
        #[arg(env = "SERVICE_HOST", long, default_value = "127.0.0.1:3000")]
        host: SocketAddr,
        #[arg(env = "GRPC_SERVICE_HOST", long, default_value = "127.0.0.1:7900")]
        grpc_host: SocketAddr,
    },

    Client {
        #[arg(env = "SERVICE_HOST", long, default_value = "http://127.0.0.1:3000")]
        host: String,
        #[arg(
            env = "GRPC_SERVICE_HOST",
            long,
            default_value = "http://127.0.0.1:7900"
        )]
        grpc_host: String,

        #[command(subcommand)]
        commands: ClientCommands,
    },
}

#[derive(Subcommand)]
enum ClientCommands {
    PublishEvent {
        #[arg(long)]
        topic: String,
        #[arg(long)]
        key: String,
        #[arg(long)]
        value: String,
        #[arg(long)]
        id: Option<String>,
        #[arg(long = "generate-id")]
        generate_id: bool,
    },
    GetTopics {},
    GetKeys {
        #[arg(long)]
        topic: String,
    },
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    dotenv::dotenv().ok();
    tracing_subscriber::fmt::init();

    let cli = Command::parse();
    match cli.command.unwrap() {
        Commands::Serve { host, grpc_host } => {
            tracing::info!("Starting service");
            let state = SharedState::new().await?;
            Mad::builder()
                .add(HttpServer::new(&state, host))
                .add(GrpcServer::new(&state, grpc_host))
                .run()
                .await?;
        }
        Commands::Client {
            commands,
            host,
            grpc_host,
        } => match commands {
            ClientCommands::PublishEvent {
                topic,
                key,
                value,
                id,
                generate_id,
            } => {
                let mut client = create_client(grpc_host).await?;

                let timestamp = chrono::Utc::now();

                let _ = client
                    .publish_event(PublishEventRequest {
                        topic,
                        published: Some(prost_types::Timestamp::date_time_nanos(
                            timestamp.year() as i64,
                            timestamp.month() as u8,
                            timestamp.day() as u8,
                            timestamp.hour() as u8,
                            timestamp.minute() as u8,
                            timestamp.second() as u8,
                            timestamp.nanosecond(),
                        )?),
                        key,
                        value: value.into_bytes(),
                        id: {
                            if id.is_none() {
                                if generate_id {
                                    Some(Uuid::new_v4().to_string())
                                } else {
                                    None
                                }
                            } else {
                                id
                            }
                        },
                    })
                    .await?;
            }
            ClientCommands::GetTopics {} => {
                let mut client = create_client(grpc_host).await?;

                println!("Listing topics");
                let topics = client.get_topics(GetTopicsRequest {}).await?;

                for topic in topics.into_inner().topics {
                    println!("{topic}");
                }
            }
            ClientCommands::GetKeys { topic } => {
                let mut client = create_client(grpc_host).await?;

                println!("Listing keys for topic: {}", topic);
                let keys = client.get_keys(GetKeysRequest { topic }).await?;

                for key in keys.into_inner().keys {
                    println!("{key}");
                }
            }
        },
    }

    Ok(())
}

async fn create_client(
    grpc_host: String,
) -> anyhow::Result<crate::grpc::no_data_client::NoDataClient<tonic::transport::Channel>> {
    let client = crate::grpc::no_data_client::NoDataClient::connect(grpc_host).await?;

    Ok(client)
}
