use std::{
    collections::BTreeMap,
    sync::{Arc, RwLock},
};

use crate::state::SharedState;

pub struct StagingEvent {
    pub topic: String,
    pub key: String,
    pub id: Option<String>,
}

#[derive(Default, Clone)]
pub struct Staging {
    // Temporary until we've got an actual file disk store
    store: Arc<RwLock<BTreeMap<String, BTreeMap<String, Vec<StagingEvent>>>>>,
}

impl Staging {
    pub async fn publish(&self, staging_event: impl Into<StagingEvent>) -> anyhow::Result<()> {
        let staging_event: StagingEvent = staging_event.into();
        let mut store = self.store.write().unwrap();
        tracing::trace!(
            topic = staging_event.topic,
            id = staging_event.id,
            "moving event to staging"
        );

        match store.get_mut(&staging_event.topic) {
            Some(part) => match part.get_mut(&staging_event.key) {
                Some(existing_key_part) => {
                    if staging_event.id.is_none()
                        || !existing_key_part.iter().any(|p| p.id == staging_event.id)
                    {
                        existing_key_part.push(staging_event);
                    } else {
                        tracing::debug!(
                            topic = staging_event.topic,
                            id = staging_event.id,
                            "event already found, skipping"
                        );
                    }
                }
                None => {
                    part.insert(staging_event.key.to_owned(), vec![staging_event]);
                }
            },
            None => {
                tracing::debug!(
                    topic = staging_event.topic,
                    id = staging_event.id,
                    "new topic, creating partition"
                );
                store.insert(
                    staging_event.topic.to_owned(),
                    BTreeMap::from([(staging_event.key.to_owned(), vec![staging_event])]),
                );
            }
        }

        Ok(())
    }

    pub async fn get_topics(&self) -> anyhow::Result<Vec<String>> {
        let store = self.store.read().unwrap();

        Ok(store.keys().cloned().collect::<Vec<_>>())
    }

    pub async fn get_keys(&self, topic: impl Into<String>) -> anyhow::Result<Vec<String>> {
        let store = self.store.read().unwrap();

        let items = store
            .get(&topic.into())
            .map(|tree| tree.keys())
            .unwrap_or_default();

        Ok(items.cloned().collect::<Vec<_>>())
    }
}

pub trait StagingState {
    fn staging(&self) -> Staging;
}

impl StagingState for SharedState {
    fn staging(&self) -> Staging {
        self.staging.clone()
    }
}
